"""
The OBI-WAN estimator proposed in the paper 
"A Permutation-based Model for Crowd Labeling: Optimal Estimation and Robustness" 
by Nihar B. Shah, Sivaraman Balakrishnan and Martin J. Wainwright

Inputs:
 - data matrix Y, which is an (n x d) matrix with values in {-1,0,1} where n = number of workers and d = number of questions
 - threshold: a positive number, where the windowing step of the constituent of the WAN estimator "accepts" a bias of numpy.sqrt(k) *threshold) toward any answer; default value is numpy.sqrt((numpy.log(d*n))**1.5)*p

Output:
 - vector xhat which is an n-length vector taking values in {-1,1}, which is an estimate of the labels
 """

import numpy
from estimator_WAN import *

def estimator_OBIWAN(Y, threshold = -1):
    n = numpy.size(Y, 0)
    d = numpy.size(Y, 1)
    xhat = numpy.zeros((d))

    #Divide the questions into two bins
    b = numpy.random.binomial(1, 0.5, d)
    Y1 = Y[:, numpy.nonzero(1-b)[0]]
    Y2 = Y[:, numpy.nonzero(b)[0]]
    
    U, s, V = numpy.linalg.svd(Y1, full_matrices = False)
    u1 = U[:,0]
    #fix sign of u1
    u1 = u1* numpy.sign( numpy.sum( u1*u1*numpy.sign(u1) ) )
    
    U, s, V = numpy.linalg.svd(Y1, full_matrices = False)
    u2 = U[:,0]
    #fix sign of u2
    u2 = u2* numpy.sign( numpy.sum( u2*u2*numpy.sign(u2) ) )

    #use the WAN estimator with orderings given by u1 and u2    
    xhat1 = estimator_WAN(Y1, permwork = u2, threshold = threshold)
    xhat2 = estimator_WAN(Y2, permwork = u1, threshold = threshold)
    
    #Put the two estimates back together
    j1 = 0
    j2 = 0
    for j in numpy.arange(d):
        if b[j] == 0:
            xhat[j] = xhat1[j1]
            j1+=1
        else:
            xhat[j] = xhat2[j2]
            j2+=1
    
    return xhat