"""
The OBI-WAN estimator proposed in the paper 
"A Permutation-based Model for Crowd Labeling: Optimal Estimation and Robustness" 
by Nihar B. Shah, Sivaraman Balakrishnan and Martin J. Wainwright

Inputs:
 - data matrix Y, which is an (n x d) matrix with values in {-1,0,1} where n = number of workers and d = number of questions
 - permutation permwork of the workers which is an (approximate) ordering of the workers in terms of their abilities
 - threshold: a positive number, where the windowing step "accepts" a bias of numpy.sqrt(k) *threshold) toward any answer; default value is numpy.sqrt((numpy.log(d*n))**1.5)*p

Output:
 - vector xhat which is an n-length vector taking values in {-1,1}, which is an estimate of the labels
"""

import numpy

def estimator_WAN(Y, permwork, threshold = -1):
    n = numpy.size(Y, 0)
    d = numpy.size(Y, 1)
    xhat = numpy.zeros((d))
    
    Y_ord = numpy.zeros((n, d))
    
    #Estimate the probability of observing a answer-question pair
    p = numpy.sum(numpy.abs(Y))*1.0/(n*d)
    
    if threshold == -1:
        threshold = numpy.sqrt((numpy.log(d*n))**1.5)*p
    
    #Order the rows of matrix Y according to the order specified by entries of permwork
    permwork = numpy.argsort(-permwork)

    for i in numpy.arange(n):
        Y_ord[i,:] = Y[permwork[i],:]

    #Windowing: find "k" such that top k workers are good        
    k = n
    numBiased = 0 #number of questions for which bias is more than the given threshold
    majVote = numpy.sum(Y_ord, axis = 0) #majority vote of all workers
    for ktemp in numpy.arange(n, 0, -1):
        if numpy.count_nonzero( numpy.abs(majVote) >= numpy.sqrt(ktemp) *threshold) > numBiased: #see if majority vote of first ktemp workers has enough bias
            numBiased =  numpy.count_nonzero( numpy.abs(majVote) >= numpy.sqrt(ktemp) * threshold) 
            k = ktemp
        majVote -= Y_ord[ktemp-1,:] #take majority vote of first k-1 workers
        
        
    #Aggregating Naively: majority vote of top k workers
    xhat = numpy.sign( numpy.sum(Y_ord[0:k, :], axis = 0) )
    return xhat